@@ -4,19 +4,23 @@
  * Redistribution and use in source and binary forms are permitted
  * provided that this notice is preserved and due credit is given
  * to the original author and the contributors.
- * $Id: ipt.h,v 2.0.2.7.2.1 1998/11/22 01:50:38 darrenr Exp $
+ * $Id: ipt.h,v 2.1 1999/08/04 17:30:08 darrenr Exp $
  */
 
 #ifndef	__IPT_H__
 #define	__IPT_H__
 
-#include <fcntl.h>
-#ifdef	__STDC__
-#include <stdarg.h>
-#else
-#include <varargs.h>
+#ifndef	__P
+# define P_DEF
+# ifdef	__STDC__
+#  define	__P(x) x
+# else
+#  define	__P(x) ()
+# endif
 #endif
 
+#include <fcntl.h>
+
 
 struct	ipread	{
 	int	(*r_open) __P((char *));
@@ -27,4 +31,9 @@ struct	ipread	{
 extern	void	debug __P((char *, ...));
 extern	void	verbose __P((char *, ...));
 
+#ifdef P_DEF
+# undef	__P
+# undef	P_DEF
+#endif
+
 #endif /* __IPT_H__ */
